#-*- coding:utf-8 -*-

#  Pybik -- A 3 dimensional magic cube game.
#  Copyright (C) 2009  B. Clausius
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License,  or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


# Ported from GNUbik
# Original filename: cube.c
# Original copyright and license:
#/*
#  Copyright (c) 2004        Dale Mellor,  John Darrington
#  copyright (c) 1998,  2003  John Darrington
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License,  or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#*/


#px +DEBUG=False

global __name__
if DEBUG:
    print 'Importing module:', __name__


import random
rand = random.Random()
if DEBUG:
    print 'rand.seed(123)'
    rand.seed(123)

import txfm


#px/DEF MATRIX_DIM = 4
MATRIX_DIM = 4

# Unique identifiers for the faces,  which can be used to populate a
# bit-field.
#px/cdef enum:
if True:
    FACE_0 = 0x01 << 0
    FACE_1 = 0x01 << 1
    FACE_2 = 0x01 << 2
    FACE_3 = 0x01 << 3
    FACE_4 = 0x01 << 4
    FACE_5 = 0x01 << 5


# Is the cube solved?
#typedef enum _Cube_Status { NOT_SOLVED=0,  SOLVED,  HALF_SOLVED } Cube_Status;
NOT_SOLVED = 0
SOLVED = 1
HALF_SOLVED = 2

#px/ctypedef float Vector[4]
Vector = txfm.Vector
#px/ctypedef float Point[4]
Point = txfm.Point
#px/ctypedef Vector Matrix[4]
Matrix = txfm.ZeroMatrix

#cdef struct Face:
class Face:
    #px+Vector quadrants[4]
    #px+Vector normal
    #px+Point centre
    #px-
    def __init__(self):
        # An array of vectors showing the orientation of the blocks. There are four
        # vectors per face. These are used to orientate the mouse pointer,  to
        # detect when the cube has been solved and to provide feedback to clients
        # querying the state.
        #px-
        self.quadrants = []  # vector[4]
        #px-
        for __i in xrange(4):
            #px-
            self.quadrants.append(txfm.Vector())
        # The normal vector is orthogonal to the face of the block. When all
        # normals are in the same direction,  the cube colours are correct,  but not
        # necessarily their orientations.
        #px-
        self.normal = txfm.Vector()  # Vector
        # The position of the centre of the face,  relative to the centre of the
        # block.
        #px-
        self.centre = txfm.Point()  # point


#cdef struct Block:
class Block:
    #px-
    def __init__(self):
        # Bit-field indicating which faces are on the surface of the cube,  and
        # should therefore be rendered to the framebuffer.
        #px/int visible_faces
        self.visible_faces = 0  # int
        # A set of attributes for each face (including internal ones!)
        #px/Face face[6]
        self.face = [Face() for __i in xrange(6)]
        # The position from the centre of the cube,  and the rotation from the
        # 'natural' position (note that the location vector is accessed as
        # transformation+12).
        #px/Matrix transformation
        self.transformation = Matrix()  # Matrix
        #px/int in_motion
        self.in_motion = False


#px/DEF MAX_BLOCKS = 10**3
MAX_BLOCKS = 10**3

#cdef struct _Cube:
class _cube:
    #px/unsigned int dimension
    dimension = None
    #px/unsigned int dimension2
    dimension2 = None
    #px/unsigned int number_blocks
    number_blocks = None
    #px/Block blocks[MAX_BLOCKS]
    blocks = None
    
    #px/int animation_running
    animation_running = 0
    #px/int move_slice
    move_slice = 0
    #px/int move_dir
    move_dir = 0
    #px/int move_axis
    move_axis = 0
    #px/float animation_angle
    animation_angle = 0
#px+cdef _Cube _cube

def get_cube_dimension():
    return _cube.dimension

#@debug
def set_cube_dimension(cube_size):
    #px+cdef Block* block
    #px+cdef int i
    
    if cube_size == 0:
        return
    
    # The number of blocks per side of the cube.
    _cube.dimension = cube_size
    _cube.dimension2 = _cube.dimension * _cube.dimension
    # _cube.dimension ** 3
    _cube.number_blocks = _cube.dimension2 * _cube.dimension
    assert _cube.number_blocks <= MAX_BLOCKS
    
    # A set of attributes for every block (including internal ones!)
    #px-
    _cube.blocks = []
    #px-
    for i in xrange(_cube.number_blocks):
    #px-
        _cube.blocks.append(Block())
    
    # Loop over the array of blocks,  and initialize each one.
    #px/for 0 <= i < _cube.number_blocks:
    for i,block in enumerate(_cube.blocks):
        #px+block = &_cube.blocks[i]
        
        # Flagging only certain faces as visible allows us to avoid rendering
        # invisible surfaces,  thus slowing down animation.
        block.visible_faces = int(
                       FACE_0 * (0 == i / _cube.dimension2)
                     + FACE_1 * (_cube.dimension - 1 == i / _cube.dimension2)
                     + FACE_2 * (0 == i / _cube.dimension % _cube.dimension)
                     + FACE_3 * (_cube.dimension - 1 == i / _cube.dimension % _cube.dimension)
                     + FACE_4 * (0 == i % _cube.dimension)
                     + FACE_5 * (_cube.dimension - 1 == i % _cube.dimension)
                     )
        
        # Initialize all transformations to the identity matrix,  then set the
        # translation part to correspond to the initial position of the
        # block.
        for ky in xrange(3):
            for kx in xrange(4):
                block.transformation[ky][kx] = 0.0
        for kx in xrange(4):
            block.transformation[kx][kx] = 1.0

        block.transformation[3][0] = block_index_to_coords (i % _cube.dimension)
        block.transformation[3][1] = block_index_to_coords ((i / _cube.dimension) % _cube.dimension)
        block.transformation[3][2] = block_index_to_coords ((i / _cube.dimension2) % _cube.dimension)

        # Set all the face centres.
        #px-
        block.face[0].centre = [ 0, 0,-1, 0]
        #px+block.face[0].centre[0] = 0.0
        #px+block.face[0].centre[1] = 0.0
        #px+block.face[0].centre[2] = -1.0
        #px+block.face[0].centre[3] = 0.0
        #px-
        block.face[1].centre = [ 0, 0, 1, 0]
        #px+block.face[1].centre[0] = 0.0
        #px+block.face[1].centre[1] = 0.0
        #px+block.face[1].centre[2] = 1.0
        #px+block.face[1].centre[3] = 0.0
        #px-
        block.face[2].centre = [ 0,-1, 0, 0]
        #px+block.face[2].centre[0] = 0.0
        #px+block.face[2].centre[1] = -1.0
        #px+block.face[2].centre[2] = 0.0
        #px+block.face[2].centre[3] = 0.0
        #px-
        block.face[3].centre = [ 0, 1, 0, 0]
        #px+block.face[3].centre[0] = 0.0
        #px+block.face[3].centre[1] = 1.0
        #px+block.face[3].centre[2] = 0.0
        #px+block.face[3].centre[3] = 0.0
        #px-
        block.face[4].centre = [-1, 0, 0, 0]
        #px+block.face[4].centre[0] = -1.0
        #px+block.face[4].centre[1] = 0.0
        #px+block.face[4].centre[2] = 0.0
        #px+block.face[4].centre[3] = 0.0
        #px-
        block.face[5].centre = [ 1, 0, 0, 0]
        #px+block.face[5].centre[0] = 1.0
        #px+block.face[5].centre[1] = 0.0
        #px+block.face[5].centre[2] = 0.0
        #px+block.face[5].centre[3] = 0.0
        
        block.in_motion = 0
        
    _cube.animation_running = 0
        
# Cube co-ordinates have their origin at the centre of the cube,  and their
# units are equivalent to one half of the length of one edge of a block.
def block_index_to_coords(i):
    return float(2 * i - (_cube.dimension - 1))

def block_coords_to_index(i):
    return int((i + (_cube.dimension - 1)) / 2.0)

# Utility function to fetch a particular face of the cube.
#cdef Face* get_face(int block, int face):
def get_face(block, face):
    #px/return &_cube.blocks[block].face[face]
    return _cube.blocks[block].face[face]


# Structure used to communicate which blocks of the cube lie in a certain
# slice. These objects should only be constructed by one of the identify_blocks
# functions.
#FIXME: remove
class Slice_Blocks:
    def __init__(self):
        self.blocks = []        # int[]
        #self.number_blocks = 0  # uint
        self.axis = 0           # uint

def cube_randomize():
    for i in xrange(8*_cube.dimension):
        blocks = identify_blocks(rand.randrange(_cube.number_blocks), rand.randrange(3))
        rotate_slice (rand.randrange(2)+1, blocks)



#   Cube co-ordinates have their origin at the centre of the cube,  and their
#   units are equivalent to one half of the length of one edge of a block.
#   This func initialises the positions of the blocks which comprise the cube.
#   The enumeration scheme I have chosen goes around four surfaces of the cube,
#   then fills in the ends.  Thus,  for a 4x4x4 cube it looks like:
#
#   ----------------------------------------------------------------------------
#   View this diagram with 132 coloumns!
#
#   |  60 |  61 |  62 |  63 |
#   |  44 |  45 |  46 |  47 |    |  56 |  57 |  58 |  59 |
#   |  28 |  29 |  30 |  31 |    |  40 |  41 |  42 |  43 |    |  52 |  53 |  54 |  55 |
#   |  12 |  13 |  14 |  15 |    |  24 |  25 |  26 |  27 |    |  36 |  37 |  38 |  39 |    |  48 |  49 |  50 |  51 |
#   |   8 |   9 |  10 |  11 |    |  20 |  21 |  22 |  23 |    |  32 |  33 |  34 |  35 |
#   |   4 |   5 |   6 |   7 |    |  16 |  17 |  18 |  19 |
#   |   0 |   1 |   2 |   3 |


# Quick cosine function for angles expressed in quarters of complete
# revolutions.
def cos_quadrant(quarters):
    quarters = abs(quarters) % 4
    if quarters == 0:
        return 1
    elif quarters == 2:
        return -1
    else:
        return 0

# Quick sine function,  for angles expressed in quarters of complete
# revolutions.
def sin_quadrant(quarters):
    return cos_quadrant(quarters - 1)

# Pre-multiply a matrix N,  by matrix M
#cdef pre_mult (Vector* M, Vector* N):
def pre_mult (M, N):
    # Temporary Matrix variable for result
    #px/cdef Matrix T
    T = Matrix()

    for k in xrange(MATRIX_DIM):
        for i in xrange(MATRIX_DIM):
            T[k][i] = 0.0
            for j in xrange(MATRIX_DIM):
                 T[k][i] = T[k][i] + M[j][i] * N[k][j]
    
    # Now copy  T into N
    for i in xrange(MATRIX_DIM):
        for j in xrange(MATRIX_DIM):
            N[i][j] = T[i][j]


def IdentityMatrix():
    return [    [1, 0, 0, 0],
                [0, 1, 0, 0],
                [0, 0, 1, 0],
                [0, 0, 0, 1],
            ]
#cdef set_matrix_identity(Matrix M):
def set_matrix_identity(M):
    M[0][0] = 1; M[0][1] = 0; M[0][2] = 0; M[0][3] = 0
    M[1][0] = 0; M[1][1] = 1; M[1][2] = 0; M[1][3] = 0
    M[2][0] = 0; M[2][1] = 0; M[2][2] = 1; M[2][3] = 0
    M[3][0] = 0; M[3][1] = 0; M[3][2] = 0; M[3][3] = 1


# Rotate the slice identified by a prior call to identify_blocks,  about the
# axis,  through an angle specified by turns,  which is in quarters of complete
# revolutions.
def rotate_slice(turns, slice_):
    #px+cdef Matrix rotation

    # Create a matrix describing the rotation of we are about to perform.
    # We do this by starting with the identity matrix...
    #px/set_matrix_identity(rotation)
    rotation = IdentityMatrix()
    
    # ... and then assigning values to the active elements.
    rotation [(slice_.axis + 1) % 3][(slice_.axis + 1) % 3] = cos_quadrant (turns)
    rotation [(slice_.axis + 2) % 3][(slice_.axis + 2) % 3] = cos_quadrant (turns)
    rotation [(slice_.axis + 2) % 3][(slice_.axis + 1) % 3] = sin_quadrant (turns)
    rotation [(slice_.axis + 1) % 3][(slice_.axis + 2) % 3] = -sin_quadrant (turns)

    # Apply the rotation matrix to all the blocks in this slice. We iterate
    # backwards to avoid recalculating the end of the loop with every
    # iteration.
    for block in reversed(slice_.blocks):
        #px/pre_mult(&rotation[0], &_cube.blocks[block].transformation[0])
        pre_mult(rotation, _cube.blocks[block].transformation)


#cdef rotate_slice_3(turns, axis):
def rotate_slice_3(turns, axis):
    #px+cdef int i
    #px+cdef Matrix rotation

    # Create a matrix describing the rotation of we are about to perform.
    # We do this by starting with the identity matrix...
    #px/set_matrix_identity(rotation)
    rotation = IdentityMatrix()
    
    # ... and then assigning values to the active elements.

    rotation [(axis + 1) % 3][(axis + 1) % 3] = cos_quadrant (turns)
    rotation [(axis + 2) % 3][(axis + 2) % 3] = cos_quadrant (turns)
    rotation [(axis + 2) % 3][(axis + 1) % 3] = sin_quadrant (turns)
    rotation [(axis + 1) % 3][(axis + 2) % 3] = -sin_quadrant (turns)

    # Apply the rotation matrix to all the blocks in this slice. We iterate
    # backwards to avoid recalculating the end of the loop with every
    # iteration.
    #px/for _cube.number_blocks > i >= 0:
    for i in reversed(xrange(_cube.number_blocks)):
        if _cube.blocks[i].in_motion:
            #px/pre_mult(&rotation[0], &_cube.blocks[i].transformation[0])
            pre_mult(rotation, _cube.blocks[i].transformation)


#********************
#* Return a pointer to an array of block numbers which are in the slice
#* identified by slice_depth (two publicly-exposed wrappers defined below allow
#* for a convenient means of computing this),  and axis (an integer in [0,
#* 3)).
#******
#FIXME use identify_blocks_3
def identify_blocks_2(slice_depth, axis):
    ret = Slice_Blocks()
    ret.blocks = []

    # We need to pass the axis on to the rotate routine.
    ret.axis = axis

    # Iterate over all the blocks in the cube. When we find one whose
    # axis-component of the location part of its transformation matrix
    # corresponds to the requested slice depth,  then we make a note of its
    # offset in the return->blocks array.
    for i in reversed(xrange(_cube.number_blocks)):
        if abs(_cube.blocks[i].transformation[3][axis] - slice_depth) < 0.1:
            ret.blocks.append(i)

    return ret

#px/cdef int identify_blocks_3(int slice_depth, int axis):
def identify_blocks_3(slice_depth, axis):
    #px+cdef ret
    ret = 0
    # Iterate over all the blocks in the cube. When we find one whose
    # axis-component of the location part of its transformation matrix
    # corresponds to the requested slice depth,  then we make a note
    for i in reversed(xrange(_cube.number_blocks)):
        if abs(_cube.blocks[i].transformation[3][axis] - slice_depth) < 0.1:
            _cube.blocks[i].in_motion = 1
            ret = 1
        else:
            _cube.blocks[i].in_motion = 0
    return ret


# Get the blocks in the same slice as the block with block_id.

def identify_blocks(block_id, axis):
    return identify_blocks_2(
                _cube.blocks[block_id].transformation[3][axis],
                axis)

def rotate_back(move_data):
    identify_blocks_3 (move_data.slice, move_data.axis)
    if move_data.dir == 0:
        rotate_slice_3 (1, move_data.axis)
    else:
        rotate_slice_3 (3, move_data.axis)
    

#********************
#* Return the quadrant vector in v.
#******/

quadrant_vector = None
def get_quadrant_vector(block, face, quadrant):
    #return get_face(block, face).quadrants[quadrant]
    #px+cdef float *v
    #px/v = &get_face(block, face).quadrants[quadrant][0]
    v = get_face(block, face).quadrants[quadrant]
    global quadrant_vector
    quadrant_vector = [v[0], v[1], v[2], v[3]]
    return quadrant_vector


#cdef vectors_equal (float *v1, float *v2):
def vectors_equal(v1, v2):
    #px+cdef int i
    #px/for 0 <= i < MATRIX_DIM:
    for i in xrange(MATRIX_DIM):
        if v1[i] != v2[i]:
            return 0
    return 1


#********************
#* The cube is solved iff for all faces the quadrant vectors point in the same
#* direction.  If however all the normals point in the same direction,  but the
#* quadrants do not,  then the colours are all on the right faces,  but not
#* correctly orientated.
#******

def cube_status_check():
    # Loop variables for iterating over faces and blocks.

    #int face
    #px+cdef Block *block
    #px +cdef Vector *v0,*v1,*q0,*q1
    #px+cdef float *v0,*v1,*q0,*q1


    # Find out if the cube is at least half solved (the colours are right,  but
    # some orientations are wrong (this can be seen on a face away from an edge
    # of the cube with a pixmap on it). If the cube is not at least
    # half-solved,  then it is definitely unsolved and this value is
    # returned.

    for face in xrange(6):
        #q0 = txfm.Vector()
        mask = 0x01 << face
        x = 0

        #px/for _cube.number_blocks > i >= 0:
        for block in reversed(_cube.blocks):
            #px+block = &_cube.blocks[i]
            if block.visible_faces & mask:
                v0 = block.face[face].normal
                if x == 0:
                    q0 = v0
                    x = x+1
                else:
                    if not vectors_equal(q0, v0):
                        return NOT_SOLVED


    # The cube is at least half-solved. Check if it is fully solved by checking
    # the alignments of all the quadrant vectors. If any are out,  then return
    # the half-solved status to the caller. Note that it is only necessary to
    # check two perpendicular quadrant vectors.

    for face in xrange(6):
        #q1 = txfm.Vector()
        #q0 = txfm.Vector()
        mask = 0x01 << face
        x = 0

        #px/for _cube.number_blocks > i >= 0:
        for block in reversed(_cube.blocks):
            #px+block = &_cube.blocks[i]
            # Ignore faces which are inside the cube.
            if block.visible_faces & mask:
                v0 = block.face[face].quadrants[0]
                v1 = block.face[face].quadrants[1]

                if x == 0:
                    q0 = v0#[:]
                    q1 = v1#[:]
                    x = x+1
                elif not vectors_equal(q0, v0) or not vectors_equal(q1, v1):
                    return HALF_SOLVED

    # Cube is fully solved.
    return SOLVED


#********************
#* Cube accessor method.
#******

#cdef int get_visible_faces(int block_id):
def get_visible_faces(block_id):
    return _cube.blocks[block_id].visible_faces


#cdef print_matrix(Matrix m):
def print_matrix(m):
    for i_ in xrange(4):
        for j_ in xrange(4):
            print '%9.4f' % m[i_][j_],
        print


#********************
#* Get the transformation of block number `block_id' from the origin,  and store
#* it in transform. Return 0 on success,  1 on error.
#******

M = []
def get_block_transform(block_id):
    #px+cdef Vector *T
    #px+cdef int i,j
    #px+T = &_cube.blocks[block_id].transformation[0]
    #print_matrix(T)
    #px+global M
    #px/M = []
    M = _cube.blocks[block_id].transformation#[:]
    #px+for 0 <= j < MATRIX_DIM:
    #px+    v = []
    #px+    for 0 <= i < MATRIX_DIM:
    #px+        v.append(T[j][i])
    #px+    M.append(v)
    #print M
    return M
# cdef Matrix _get_block_transform(int block_id):
#cdef float* _get_block_transform(int block_id):
def _get_block_transform(block_id):
    #px/return &_cube.blocks[block_id].transformation[0][0]
    return _cube.blocks[block_id].transformation#[:]


#********************
def get_pending_movement_slice(selection_block, pending_movement_axis):
    return _cube.blocks[selection_block].transformation[3][pending_movement_axis]


# Pre-multiply a point or vector x,  by matrix M
#cdef void transform(Vector q, Matrix M, Vector x):
def transform(q, M, x):
    #px+cdef int i
    #px-
    assert len(q) == 0

    # for each row
    for i in xrange(MATRIX_DIM):
        f = 0
        # for each column
        for j in xrange(MATRIX_DIM):
            f = f + M[j][i] * x[j]
        #px/q[i] = f
        q.append(f)


#********************
#* Manufacture a object which is a vector of six vectors of (cube_size *
#* cube_size) elements,  each one holding the colour of a patch of the surface of
#* the cube (a number from [0, 5]).
#********
# ((1,3,size,size,size), colors[6][size*size])
#

def make_plugin_cube():
    #px+cdef Block *block
    #px+cdef int i
    #px+cdef float *face_direction_
    #px+cdef Vector face_direction
    #int face
    #char colours [6] [_cube.dimension] [_cube.dimension]
    colours = []
    for __j in xrange(6):
        colour = []
        for __i in xrange(_cube.dimension):
            colour.append(list(' '*_cube.dimension))
        colours.append(colour)

    # Loop over all blocks and faces,  but only process if the face is on the
    # outside of the cube.

    #px/for _cube.number_blocks > i >= 0:
    for block in reversed(_cube.blocks):
        #px+block = &_cube.blocks[i]
        for face in xrange(6):
            if block.visible_faces & (0x01 << face):
                # Apply the rotation part of the block transformation to the
                # offset of the face centre from the centre of the block. This
                # will tell us the direction the face is now facing.

                #point face_direction

                #px-
                face_direction = []
                face_direction_ = block.face[face].centre#[:]
                transform (face_direction, block.transformation,  face_direction_)


                # The face direction will have exactly one non-zero component
                # this is in the direction of the normal,  so we can infer which
                # side of the cube we are on. Further,  if we look at the other
                # two components of the location part of the block
                # transformation,  we can infer the position of the block in the
                # face. We set the appropriate element in the colours array to
                # the colour of this face,  which corresponds exactly with the
                # original face the block was at.

                if abs(face_direction[0]) > 0.1:
                    if face_direction[0] < 0.0: f = 4
                    else:                       f = 5
                    colours[f][
                            block_coords_to_index(block.transformation[3][1])][
                            block_coords_to_index(block.transformation[3][2])] = face

                elif abs(face_direction[1]) > 0.1:
                    if face_direction[1] < 0.0: f = 2
                    else:                       f = 3
                    colours[f][
                            block_coords_to_index(block.transformation[3][0])][
                            block_coords_to_index(block.transformation[3][2])] = face

                else:
                    if face_direction[2] < 0.0: f = 0
                    else:                       f = 1
                    colours[f][
                            block_coords_to_index(block.transformation[3][0])][
                            block_coords_to_index(block.transformation[3][1])] = face

    # Now the colours array should be completely populated
    
    #face_vector = [None]*6
    face_vector = []

    for face in xrange(6):
        #block_vector = [None]*(_cube.dimension * _cube.dimension)
        block_vector = []
        for j in xrange(_cube.dimension):
            for i in xrange(_cube.dimension):
                block_vector.append(colours[face][i][j])
        face_vector.append(block_vector)
        #face_vector.append([colours[face][i][j] for j in xrange(_cube.dimension)
        #                                            for i in xrange(_cube.dimension)])

    return ((1, 3, _cube.dimension, _cube.dimension, _cube.dimension), face_vector)

