#-*- coding:utf-8 -*-

#  Pybik -- A 3 dimensional magic cube game.
#  Copyright (C) 2009  B. Clausius
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License,  or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


# Ported from GNUbik
# Original filename: glarea-common.c
# Original copyright and license:
#/*
#    GNUbik -- A 3 dimensional magic cube game.
#    Copyright (C) 2003  John Darrington
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License,  or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#*/


import quarternion
from OpenGL import GL as gl
from OpenGL import GLU as glu
import math
#import glarea


#import txfm
import cube

fovy = -1
cp_near = -1
cp_far = -1


# Start with the unit quarternion
cube_view = quarternion.Quarternion(1,  0,  0, 0)

bounding_sphere_radius = 0

#struct jitter_v {
#  GLdouble x
#  GLdouble y
#}
class jitter_v:
    def __init__(self, x=0, y=0):
        self.x = x
        self.y = y


j8 = [
         jitter_v(0.5625, 0.4375),
         jitter_v(0.0625, 0.9375),
         jitter_v(0.3125, 0.6875),
         jitter_v(0.6875, 0.8125),
         jitter_v(0.8125, 0.1875),
         jitter_v(0.9375, 0.5625),
         jitter_v(0.4375, 0.0625),
         jitter_v(0.1875, 0.3125),
     ]


# Wrapper to set the modelview matrix
def modelViewInit ():
    # start with the cube slightly skew,  so we can see all its aspects
    cube_orientation = (15.0,  15.0)

    origin = (0, 0, 0)

    # Update viewer position in modelview matrix

    gl.glMatrixMode (gl.GL_MODELVIEW)
    gl.glLoadIdentity ()
    glu.gluLookAt (0, 0,  bounding_sphere_radius + cp_near,
              origin[0],  origin[1],  origin[2],   0.0,  1.0,  0.0)


    m = quarternion.quarternion_to_matrix (cube_view)
    gl.glMultMatrixf (m)

    # skew the cube
    gl.glRotatef (cube_orientation[1], 1, 0, 0) # horizontally
    gl.glRotatef (cube_orientation[0], 0, 1, 0) # vertically

    #
    # DM 3-Jan-2004
    #
    # Add a couple of 90 degree turns to get the top and right faces in their
    # logical positions when the program starts up.
    gl.glRotatef (90.0, 1, 0, 0)
    gl.glRotatef (-90.0, 0, 0, 1)


# accFrustum and accPerspective are taken from p397 of the Red Book

def accFrustum (left, right, bottom, top, zNear, zFar,
                pixdx, pixdy, eyedx, eyedy, focus):
    dx = 0
    dy = 0
    #GLdouble xwsize,  ywsize
    viewport = [0]*4

    viewport = gl.glGetIntegerv (gl.GL_VIEWPORT)

    xwsize = right - left
    ywsize = top - bottom

    #print viewport[2], viewport[3], focus
    dx = -(pixdx*xwsize/viewport[2] + eyedx*zNear/focus)
    dy = -(pixdy*ywsize/viewport[3] + eyedy*zNear/focus)

    gl.glFrustum (left+dx,  right+dy,  bottom+dy,  top+dy,  zNear,  zFar)


def accPerspective (fovy, aspect, zNear, zFar,
                    pixdx, pixdy, eyedx, eyedy, focus):
    #GLdouble fov2 ,  left,  right,  top,  bottom

    fov2 = (fovy*math.pi / 180.0) / 2.0
    top = zNear/(math.cos (fov2)/math.sin (fov2))
    bottom = -top
    right = top*aspect
    left = -right
    accFrustum (left,  right,  bottom,  top,  zNear,  zFar,
               pixdx,  pixdy,  eyedx,  eyedy,  focus)


def projection_init (jitter):
    global bounding_sphere_radius, fovy, cp_near, cp_far
    bounding_sphere_radius = 2 * cube.get_cube_dimension()

    fovy = 33.0
    cp_near = bounding_sphere_radius / ( math.tan(fovy*math.pi / 360.0) )
    cp_far = cp_near + 2*bounding_sphere_radius

    gl.glEnable (gl.GL_DEPTH_TEST)
    gl.glClearColor (0, 0, 0, 0)

    gl.glMatrixMode (gl.GL_PROJECTION)
    gl.glLoadIdentity ()

    accPerspective (fovy,  1,  cp_near,  cp_far,
                   j8[jitter].x,  j8[jitter].y,
                   0.0, 0.0, 1.0)


def lighting_init(lighting=False):
    light_ambient = [0.6, 0.6, 0.6, 1.0]
    mat_diffuse = [0.8, 0.8, 0.8, 1.0]
    mat_specular = [0.2, 0.2, 0.2, 1.0]
    mat_shininess = 1.0
    light_position = [-1.0, 0.5, 0.0, 0.0]

    gl.glShadeModel (gl.GL_SMOOTH)

    gl.glColorMaterial (gl.GL_FRONT,  gl.GL_AMBIENT_AND_DIFFUSE)

    gl.glMaterialfv (gl.GL_FRONT,  gl.GL_SPECULAR,  mat_specular)
    gl.glMaterialfv (gl.GL_FRONT,  gl.GL_DIFFUSE,  mat_diffuse)
    gl.glMaterialf (gl.GL_FRONT,  gl.GL_SHININESS,  mat_shininess)

    gl.glLightfv (gl.GL_LIGHT0,  gl.GL_POSITION,  light_position)
    gl.glLightfv (gl.GL_LIGHT1,  gl.GL_AMBIENT,  light_ambient)

    (gl.glEnable if lighting else gl.glDisable)(gl.GL_LIGHTING)

    gl.glEnable (gl.GL_COLOR_MATERIAL)

    gl.glEnable (gl.GL_LIGHT0)
    gl.glEnable (gl.GL_LIGHT1)


