#-*- coding:utf-8 -*-

#  Pybik -- A 3 dimensional magic cube game.
#  Copyright (C) 2009  B. Clausius
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License,  or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


# Ported from GNUbik
# Original filename: widget-set.c
# Original copyright and license:
#/*
#    GNUbik -- A 3 dimensional magic cube game.
#    Copyright (C) 2003, 2004  John Darrington
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License,  or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#*/


import os

import config
import dialogs
import colour_sel
import plugin_hooks
import move_queue_m
import ui
import textures

import gobject
import gtk
from gtk import gdk
#import gtk.glade
from gtk import gtkgl
from gtk import gdkgl

import drwBlock
import glarea
import cube
from confstore import confstore, COLORED, IMAGED


N_ = lambda string: string

main_application_window = None


class MainWindow (object):
    def __init__(self, argv):
        global main_application_window
        assert main_application_window == None
        
        confstore.callbacks.append(self.on_confstore_changed)
        
        #gladefile = os.path.join(config.UI_DIR, 'pybik.glade')
        gladefile = os.path.join(config.UI_DIR, 'pybik.ui')
        #self.ui_xml = gtk.glade.XML(gladefile)
        self.ui_xml = gtk.Builder()
        self.ui_xml.props.translation_domain = config.PACKAGE
        self.ui_xml.add_from_file(gladefile)
        missing = self.ui_xml.connect_signals(self)
        if DEBUG and missing:
            print 'missing handlers:'
            print '\n'.join(missing)
        
        self.window = self.get_object('window1')
        self.play_toolbar = self.get_object('toolbar1')
        self.statusbar = self.get_object('statusbar1')
        
        pixmapfile = config.IMAGE_DIR + '.png'
        self.window.set_icon_from_file(pixmapfile)
        
        self.window.connect("delete-event", self.gtk_main_quit)

        self.create_play_toolbar()
        self.create_gl_area(argv)
        self.create_sidepane()
        
        main_application_window = self.window
    
    #get_object = lambda self, obj:      self.ui_xml.get_widget(obj)
    get_object = lambda self, obj:     self.ui_xml.get_object(obj)
    #connect_signals = lambda self, cont:    self.ui_xml.signal_autoconnect(cont)
    connect_signals = lambda self, cont:   self.ui_xml.connect_signals(cont)
    
    @staticmethod
    def on_confstore_changed(key, *subkeys):
        def set_color(i):
                color = gdk.Color(confstore.colors[i].color)
                #confstore.colors[i].facetype = COLORED
                drwBlock.face_rendering_set(i, red=color.red / 65535.0,
                                               green=color.green / 65535.0,
                                               blue=color.blue / 65535.0,
                                               #facetype=COLORED,
                                               #texname=-1
                                           )
        def set_pattern(i):
            pattern = confstore.colors[i].pattern
            texname = textures.stock_pattern[pattern].texName if pattern >= 0 else -1
            #confstore.colors[i].facetype = COLORED
            drwBlock.face_rendering_set(i, texname=texname)
        def set_imagefile(i):
            imagefile = confstore.colors[i].imagefile
            #print 'set_imagefile',i,imagefile
            try:
                pixbuf = colour_sel.create_pixbuf_from_file(imagefile)
                texname = textures.create_pattern_from_pixbuf(pixbuf)
            except Exception, err:
                print _("Cannot create image from file %s: %s") % (imagefile, err)
            else:
                #print 'image tex',texname
                drwBlock.face_rendering_set(i, texname=texname)
        
        if key == 'frameQty':
            ui.frameQty = confstore.frameQty
        elif key == 'lighting':
            glarea.set_lighting(confstore.lighting)
        elif key == 'dimension':
            pass
        elif key == 'colors':
            i = int(subkeys[0])
            if subkeys[1] == 'color':
                set_color(i)
            elif subkeys[1] == 'pattern':
                set_pattern(i)
            elif subkeys[1] == 'imagefile':
                set_imagefile(i)
            elif subkeys[1] == 'facetype':
                facetype = confstore.colors[i].facetype
                if facetype == COLORED:
                    set_pattern(i)
                elif facetype == IMAGED:
                    set_imagefile(i)
                drwBlock.face_rendering_set(i, facetype=facetype)
            elif subkeys[1] == 'imagemode':
                imagemode = confstore.colors[i].imagemode
                drwBlock.face_rendering_set(i, distr=imagemode)
                #set_imagefile(i)
            else:
                print 'Unknown conf value changed:', subkeys[1]
            glarea.postRedisplay()
        else:
            print 'Unknown conf value changed:', key
        
    @debug
    def create_sidepane(self):
        # Get the menu structure asked for by the collection of plugins.
        list_ = plugin_hooks.startup_plugins() or []
        
        treeview = self.get_object('treeview1')
        treestore = self.get_object('treestore1')
        treestore = gtk.TreeStore(gobject.TYPE_STRING, gobject.TYPE_PYOBJECT)
        treeview.set_model(treestore)
        
        column = gtk.TreeViewColumn(_('Name'), gtk.CellRendererText(), text=0)
        column.set_sizing(gtk.TREE_VIEW_COLUMN_FIXED)
        column.set_fixed_width(200)
        treeview.append_column(column)
        
        # add the items
        tree_iters = {'':None}
        for func_name, func in list_:
            p = func_name.split('/',4)
            for i in xrange(1,5):
                p_i = p[:i]
                p_is = '/'.join(p_i)
                p_i_p = p[:i-1]
                p_is_p = '/'.join(p_i_p)
                name = p_i[-1]
                if p_is not in tree_iters:
                    #print '|'.join((name, p_is, func_name))
                    ti = treestore.append(tree_iters[p_is_p], [name, func if p_is==func_name else None])
                    tree_iters[p_is] = ti
        treeview.connect('row-activated', self.on_run_script)
        treeview.expand_all()
        treeview.show_all()

    @debug
    def on_run_script(self, treeview, path, view_column):
        name, func = treeview.get_model()[path]
        #print name, func
        if callable(func):
            plugin_hooks.run_script(func, self)

    #@debug
    def create_play_toolbar(self):
        play_button_list = self.play_toolbar.get_children()
        self.playbarstate = PlaybarState(self, play_button_list)
        self.playbarstate.empty()
        #self.actions_playbar = gtk.ActionGroup('actiongroup_playbar')
    
    #@debug
    def create_gl_area(self, argv):
        target = [  ("text/uri-list",  0,  colour_sel.RDRAG_FILELIST),
                    ("application/x-color",  0,  colour_sel.RDRAG_COLOUR),
                ]
        attribs = [ gdkgl.RGBA,
                    gdkgl.RED_SIZE,    1,
                    gdkgl.GREEN_SIZE,  1,
                    gdkgl.BLUE_SIZE,   1,

                    gdkgl.DOUBLEBUFFER,

                    gdkgl.DEPTH_SIZE , 1,

                    gdkgl.ACCUM_RED_SIZE,  1,
                    gdkgl.ACCUM_GREEN_SIZE,  1,
                    gdkgl.ACCUM_BLUE_SIZE,  1,

                    gdkgl.ATTRIB_LIST_NONE
                ]
        
        glarea.glarea_init(argv)
        glconfig = gdkgl.Config (attribs)

        if not glconfig:
            #g_error ("Cannot get a suitable visual")
            raise Exception("Cannot get a suitable visual")

        glxarea = self.get_object('drawingarea1')
        glarea.glwidget = glxarea
        gtkgl.widget_set_gl_capability (glxarea,
                                      glconfig,
                                      None,
                                      True,
                                      gdkgl.RGBA_TYPE)
        
        glxarea.drag_dest_set (gtk.DEST_DEFAULT_ALL, target, gdk.ACTION_COPY)
        glxarea.connect ("drag_data_received", colour_sel.drag_data_received)

        glxarea.connect ("realize",         glarea.graphics_area_init)
        glxarea.connect ("expose_event",    glarea.expose)
        glxarea.connect ("configure_event", glarea.resize)


        glxarea.add_events (
                               gdk.KEY_PRESS_MASK | gdk.BUTTON_PRESS_MASK
                               | gdk.BUTTON_RELEASE_MASK
                               # | gdk.BUTTON_MOTION_MASK 
                               | gdk.ENTER_NOTIFY_MASK | gdk.LEAVE_NOTIFY_MASK
                               | gdk.VISIBILITY_NOTIFY_MASK
                               | gdk.POINTER_MOTION_MASK )


        glxarea.connect ("key_press_event",     glarea.cube_orientate_keys)
        glxarea.connect ("key_press_event",     self.debug_keys)
        glxarea.connect ("motion_notify_event", glarea.cube_orientate_mouse)
        glxarea.connect ("scroll_event",        glarea.z_rotate)
        glxarea.connect ("button_press_event",  glarea.buttons)
        glxarea.connect ("button_release_event",glarea.buttons)
        glxarea.connect ("button_press_event",  glarea.cube_controls, self)
    
    @debug
    def start_main_loop(self):
        self.window.show()
        gtk.main()
    
    def first_new_game(self, cube_dimension, solved):
        if cube_dimension is None:
            cube.set_cube_dimension(confstore.dimension)
        else:
            cube.set_cube_dimension(cube_dimension)
        
        if not solved:
            cube.cube_randomize()
        ui.update_ui_first(self)
        
    def new_game(self, new_size=None):
        if ui.is_animating():
            return True

        if new_size is not None:
            glarea.re_initialize_glarea()

            # create the cube
            if new_size <= 0:
                new_size = cube.get_cube_dimension()
            cube.set_cube_dimension(new_size)

        cube.cube_randomize()
        glarea.postRedisplay()
        ui.move_queue = move_queue_m.MoveQueue()
        ui.update_ui(self)

        return False
    
    # Request that the game be restarted
    # If data is non zero,  then all it's data will be reallocated
    @debug
    def request_new_game(self, unused_widget, new_size=None):
        ui.request_stop()
        if ui.is_animating():
            ui.request_abort()
        gobject.idle_add(self.new_game, new_size)
    
    
    # Action handlers
    def on_action_new_activate(self, *args):
        self.request_new_game(None)
    gtk_main_quit = gtk.main_quit
    def on_action_colors_activate(self, unused_widget):
        colour_sel.open_ColorSelectDialog(self)
    def on_action_preferences_activate(self, unused_widget):
        dialogs.preferencesdialog(self)
    def on_toggleaction_play_toolbar_toggled(self, widget):
        self.play_toolbar.props.visible = widget.props.active
    def on_toggleaction_statusbar_toggled(self, widget):
        self.statusbar.props.visible = widget.props.active
    def on_action_info_activate(self, unused_widget):
        dialogs.show_about(self.window)
    
    def on_action_rewind_activate(self, unused_widget):
        ui.request_queue_rewind(self)
    def on_action_previous_activate(self, unused_widget):
        ui.request_back(self)
    def on_action_stop_activate(self, unused_widget):
        ui.request_stop()
    def on_action_play_activate(self, unused_widget):
        ui.request_play(self)
    def on_action_next_activate(self, unused_widget):
        ui.request_next(self)
    def on_action_forward_activate(self, unused_widget):
        ui.request_fast_forward(self)
    def on_action_add_mark_activate(self, unused_widget):
        ui.request_mark_move_queue(self, True)
    def on_action_remove_mark_activate(self, unused_widget):
        ui.request_mark_move_queue(self, False)
    
    # Callbacks for PreferencesDialog
    def on_button_size_reset_activate(self, unused_widget):
        del confstore.dimension
    def on_button_animspeed_reset_activate(self, unused_widget):
        del confstore.frameQty
    def on_button_lighting_reset_activate(self, unused_widget):
        del confstore.lighting
    
    @debug
    def on_colour_select(self, unused_widget):
        colour_sel.open_ColorSelectDialog(self)
    def on_button_color_reset_activate(self, unused_widget):
        i = colour_sel.ColorSelectDialog.this_dlg.active_face[0]
        del confstore.colors[i].color
    def on_button_pattern_reset_activate(self, unused_widget):
        i = colour_sel.ColorSelectDialog.this_dlg.active_face[0]
        del confstore.colors[i].pattern
    def on_button_image_reset_activate(self, unused_widget):
        i = colour_sel.ColorSelectDialog.this_dlg.active_face[0]
        del confstore.colors[i].facetype
        del confstore.colors[i].imagefile
        del confstore.colors[i].imagemode

    
    #@debug
    def debug_keys (self, widget, event):
        shifted = 0
        if event.state & gdk.SHIFT_MASK:
            shifted = 1

        if event.keyval == gtk.keysyms.d:
            import code
            code.interact(local=globals())
            return True
        elif event.keyval == gtk.keysyms.t:
            self.on_colour_select(None)
            return True
        elif event.keyval == gtk.keysyms.q:
            gtk.main_quit()
            return True
        elif gtk.keysyms._0 <= event.keyval <= gtk.keysyms._9:
            if not hasattr(self, 'test_int'):
                self.test_int = 0
            context1 = self.statusbar.get_context_id("test1")
            context2 = self.statusbar.get_context_id("test2")
            if event.keyval == gtk.keysyms._1:
                self.test_int += 1
                mesg_id = self.statusbar.push(context1, 'Test %d' % self.test_int)
            elif event.keyval == gtk.keysyms._2:
                mesg_id = self.statusbar.pop(context1)
            elif event.keyval == gtk.keysyms._3:
                self.test_int += 1
                mesg_id = self.statusbar.push(context2, 'Test %d' % self.test_int)
            elif event.keyval == gtk.keysyms._4:
                mesg_id = self.statusbar.pop(context2)
            print '- context',context1,context2,'mesg_id',mesg_id or '-', 'value', self.test_int
            return True
        return True

    #@debug
    def update_statusbar_first(self):
        mesg = _("%d / %d moves") % (0,0)
        context = self.statusbar.get_context_id("move-count")
        self.statusbar.pop(context)
        self.statusbar.push(context, mesg)
    def update_statusbar(self):
        '''This function must be called before any action that change the move queue length
        to reflect total count of moves and after each single move'''
        mesg = _("%d / %d moves") % (
                   ui.move_queue.progress().current,
                   ui.move_queue.progress().total)
        context = self.statusbar.get_context_id("move-count")
        self.statusbar.pop(context)
        self.statusbar.push(context, mesg)

    #@debug
    def declare_win(self, status):
        if status == cube.SOLVED:
            mesg = ngettext("Cube solved in %d move", "Cube solved in %d moves",
                             ui.move_queue.progress().current
                             ) % ui.move_queue.progress().current
        elif status == cube.HALF_SOLVED:
            mesg = _("Cube is NOT solved! The colors are correct,  but have incorrect orientation")
        else:
            assert False, 'not_reached'
        context = self.statusbar.get_context_id("win")
        self.statusbar.pop(context)
        self.statusbar.push(context, mesg)
    #@debug
    def declare_win_first(self, status):
        if status == cube.SOLVED:
            mesg = ngettext("Cube solved in %d move", "Cube solved in %d moves", 0) % 0
        elif status == cube.HALF_SOLVED:
            mesg = _("Cube is NOT solved! The colors are correct,  but have incorrect orientation")
        else:
            assert False, 'not_reached'
        context = self.statusbar.get_context_id("win")
        self.statusbar.pop(context)
        self.statusbar.push(context, mesg)


class PlaybarState (object):
    def __init__(self, application, play_button_list):
        actions = [ 'action_rewind','action_previous',
                    'action_stop','action_play',
                    'action_next','action_forward',
                    'action_add_mark','action_remove_mark']
        actions = [application.get_object(a) for a in actions]
        self.play_button_list = actions
    @debug
    def set_toolbar_state(self, sflags, vflags, mark):
        #print 'set_toolbar_state'
        if mark:
            vflags ^= 0b11
        for a in reversed(self.play_button_list):
            #sens = sflags & 1
            #vis = vflags & 1
            #if sens != a.get_sensitive():
            a.set_sensitive(sflags & 1)
            #if vis != a.get_visible():
            a.set_visible(vflags & 1)
            sflags >>= 1
            vflags >>= 1
    def empty(self):          self.set_toolbar_state(0b00000000,0b11011101, False)
    def first(self):          self.set_toolbar_state(0b00011100,0b11011101, False)
    def mid(self, mark):      self.set_toolbar_state(0b11011111,0b11011110, mark)
    def last(self, mark):     self.set_toolbar_state(0b11000011,0b11011110, mark)
    def playing(self, mark):  self.set_toolbar_state(0b00100000,0b11101110, mark)


# Popup an error dialog box
@debug
def error_dialog (parent, format, *ap):
    message = format % ap
    
    dialog = gtk.MessageDialog (parent,
                     gtk.DIALOG_MODAL,
                     gtk.MESSAGE_ERROR,
                     gtk.BUTTONS_CLOSE,
                     message)
    
    dialog.set_transient_for (parent)
    # Destroy the dialog when the user responds to it (e.g. clicks a button)
    #dialog.connect ("response", gtk.Widget.destroy)
    #dialog.show_all ()
    dialog.run()
    dialog.destroy()


